#include "Tree.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>

static int* getArr(int size) {
  int* tab   = (int*)malloc(sizeof(int) * size);
  for (int i = 0; i < size; tab[i] = i++);
  return tab;
}

static void shuffle(int* theArr, int size) {
  srand((unsigned)time(NULL));
  int temporary, randomNum, last;
  for (last = size; last > 1; last--) {
    randomNum = rand( ) % last;
    temporary = theArr[randomNum];
    theArr[randomNum] = theArr[last - 1];
    theArr[last - 1] = temporary;
  }
}

static int* getRandomArr(int size) {
  int* tab = getArr(size);
  shuffle(tab, size);
  return tab;
}

static Tree* autoFillBST(Tree* t, int* input, int size) {
  deleteTree(t);
  t = createTree();
  for (int i = 0; i < size; i++) insertNode(t, input[i], "");
  return t;
}

static Tree* autoFillLLRBT(Tree* t, int* input, int size) {
  deleteTree(t);
  t = createTree();
  for (int i = 0; i < size; i++) linsertNode(t, input[i], "");
  return t;
}

static float* testTrees(Tree* t, int k, char* v, int n) {
  float* moy = (float*)malloc(sizeof(float)*2);

  for (int i = 0; i < 30; i++) {
    int* randInput = getRandomArr(n);

    t = autoFillBST(t, randInput, n);
    moy[0] += insertNode(t, k, v);


    t = autoFillLLRBT(t, randInput, n);
    moy[1] += linsertNode(t, k, v);

    free(randInput);
  }
  moy[0] = moy[0] / 30;
  moy[1] = moy[1] / 30;

  return moy;
}

/******************************************************************************
*------------------------------- Main program --------------------------------*
******************************************************************************/
int main() {
  printf("...::::: -  Binary Search Tree Implementation  - :::::...\n"
         "This program was NOT made to be robust, wrong inputs aren't supported.\n\n");
  
  Tree* tree = createTree();
  while (true) {
    printf("***********************************************************************\n"
           "Please choose an action:"
           "\n\t0: exit"
           "\n\t1: Compare -- Add a node in random trees of size N"
           "\n\t2: Compare -- Add a node in worst case (BST) of size N"
           "\n\t3: Display the tree (BST)"
           "\n\t4: Display the tree (LLRBT)"
           "\n\t5: Create random-filled BST"
           "\n\t6: Create random-filled LLRBT"
           "\n\nYour choice: ");

    int choice;
    scanf("%d", &choice);
    printf("--\n");

    switch(choice) {
      case 0:
        deleteTree(tree);
        return 0;
    
      case 1:
        printf("Enter the number of nodes in the random tree: ");
        int n;
        scanf("%d", &n);
        float* moy = testTrees(tree, n + 1, "", n);
        printf("Done.\n\t>>>>> %f comparisons performed (BST).\n"
               "\t>>>>> %f comparisons performed (LLRBT).\n", moy[0], moy[1]);
        free(moy);
        break;
        
      case 2:
        printf("Enter the number of nodes in the worst tree (for the BST): ");
        int n3;
        scanf("%d", &n3);
        tree = autoFillBST(tree, getArr(n3), n3);
        int nbCmp1 = insertNode(tree, n3+1, "");
        tree = autoFillLLRBT(tree, getArr(n3), n3);
        int nbCmp2 = linsertNode(tree, n3+1, "");

        printf("Done.\n\t>>>>> %d comparisons performed (BST).\n"
               "\t>>>>> %d comparisons performed (LLRBT).\n", nbCmp1, nbCmp2);
        break;

      case 3:
        printf("In-order traversal (BST):\n");
        displayTree(tree);
        break;

      case 4:
        printf("In-order traversal (LLRBT):\n");
        ldisplayTree(tree);
        break;

      case 5:
        printf("Enter the number of nodes: ");
        int n1;
        scanf("%d", &n1);
        printf("Creating tree... \t");
        int* randInput1 = getRandomArr(n1);
        tree = autoFillBST(tree, randInput1, n1);
        free(randInput1);
        printf("Done.\n\n");
        break;

      case 6:
        printf("Enter the number of nodes: ");
        int n2;
        scanf("%d", &n2);
        printf("Creating tree... \t");
        int* randInput2 = getRandomArr(n2);
        tree = autoFillLLRBT(tree, randInput2, n2);
        free(randInput2);
        printf("Done.\n\n");
        break;

      default:
        continue;
    }
  }

  return 0;
}
