#include "Tree.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <unistd.h> 


/******************************************************************************
*--------------------------------- New Types ---------------------------------*
******************************************************************************/
struct Keyval_t {
  int  key;
  char value[20];
};

struct Node_t {
  Node* lChild;
  Node* rChild;

  Keyval* kv;
};

struct Tree_t {
  Node* root;
  int   nbCmp;
};


/******************************************************************************
*--------------------------------- Functions ---------------------------------*
******************************************************************************/
static bool isEmpty(Tree* t) {
  return (t->root == NULL);
}

static void deleteSubtree(Node* n) {
  if (n->rChild != NULL)
    deleteSubtree(n->rChild);
  if (n->lChild != NULL)
    deleteSubtree(n->lChild);
  
  free(n);
}

static Node* createNode(Keyval* kv) {
  Node* newInstancePtr = (Node*)malloc(sizeof(Node));
  if (newInstancePtr == NULL)
    return NULL;

  newInstancePtr->lChild = NULL;
  newInstancePtr->rChild = NULL;
  newInstancePtr->kv = kv;

  return newInstancePtr;
}

static Node* getNodeInSubtree(Node* n, int k) {
  if (k < n->kv->key) {
    printf("key < %d...\n", n->kv->key);
    if (n->lChild != NULL)
      return getNodeInSubtree(n->lChild, k);
    else
      return NULL;
  }
  
  if (k > n->kv->key) {
    printf("key > %d...\n", n->kv->key);
    if (n->rChild != NULL)
      return getNodeInSubtree(n->rChild, k);
    else
      return NULL;
  }

  // Else k == n->key
  return n;
}

static Node* getNodeByKey(Tree* t, int k) {
  if (isEmpty(t))
    return NULL;

  return getNodeInSubtree(t->root, k);
}

static int cmpKeys(int a, int b) {
  return b - a;
}

static void ins(Tree* t, Node* subt, Keyval* kv) {
  int cmp = cmpKeys(subt->kv->key, kv->key);
  t->nbCmp++;
    if (cmp > 0) {
      if (subt->lChild == NULL) {
        subt->lChild = createNode(kv);
      }
      else
        ins(t, subt->lChild, kv);
    }
    else if (cmp < 0) {
      if (subt->rChild == NULL) {
        subt->rChild = createNode(kv);
      }
      else
        ins(t, subt->rChild, kv);
    }
    else {
      free(subt->kv);
      subt->kv = kv;
    }
}

static void inOrder(Node* n, Node* p) {
  int parKey = -1;
  if (p != NULL) parKey = p->kv->key;
  if(n != NULL) {
    inOrder(n->lChild, n);
    if      (p == NULL)      printf("Root -- Key: %d, value='%s'\n", n->kv->key, n->kv->value);
    else if (n == p->lChild) printf("Left child of '%d' -- Key: %d, value='%s'\n", parKey, n->kv->key, n->kv->value);
    else if (n == p->rChild) printf("Right child of '%d' -- Key: %d, value='%s'\n", parKey, n->kv->key, n->kv->value);
    inOrder(n->rChild, n);
  }
}


/******************************************************************************
*----------------------------- Accessible functions --------------------------*
******************************************************************************/
Tree* createTree() {
  Tree* newInstancePtr = malloc(sizeof(Tree));
  if (newInstancePtr == NULL)
    return NULL;

  newInstancePtr->root = NULL;

  return newInstancePtr;
}

void deleteTree(Tree* t) {
  if (t->root != NULL)
    deleteSubtree(t->root);
  free(t);
}

int insertNode(Tree* t, int key, char* value) {
  Keyval* kv = (Keyval*)malloc(sizeof(Keyval));
  kv->key = key;
  strncpy(kv->value, value, 20);

  t->nbCmp = 0;
  if (t->root == NULL)
    t->root = createNode(kv);
  else
    ins(t, t->root, kv);

  return t->nbCmp;
}

void displayNodeContent(Tree* t, int k) {
  Node* n = getNodeByKey(t, k);
  if (n == NULL)
    printf("No node within the tree recorded with this key.\n");
  else
    printf("Node found! Its value is '%s'.\n", n->kv->value);
}

void displayTree(Tree* t) {
  inOrder(t->root, NULL);
}