#include "Tree.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <time.h>
#include <unistd.h>  


/******************************************************************************
*--------------------------------- New Types ---------------------------------*
******************************************************************************/
struct Keyval_t {
  int  key;
  char value[20];
};

struct Node_t {
  Node* lChild;
  Node* rChild;

  bool color;

  Keyval* kv;
};

struct Tree_t {
  Node* root;
  int   nbCmp;
};


/******************************************************************************
*--------------------------------- Functions ---------------------------------*
******************************************************************************/
static bool isEmpty(Tree* t) {
  return (t->root == NULL);
}

static bool isRed(Node* n) {
    if (n == NULL) return false;
    return n->color == RED;
}

static Node* rotateLeft(Node* n) {
  Node* x = n->rChild;
  n->rChild = x->lChild;
  x->lChild = n;
  x->color = n->color;
  n->color = RED;
  return x;
}

static Node* rotateRight(Node* n) {
  Node* x = n->lChild;
  n->lChild = x->rChild;
  x->rChild = n;
  x->color = n->color;
  n->color = RED;
  return x;
}

static Node* flipColors(Node* n) {
  n->color = !(n->color);
  n->lChild->color = !(n->lChild->color);
  n->rChild->color = !(n->rChild->color);

  return n;
}

static void deleteSubtree(Node* n) {
  if (n->rChild != NULL)
    deleteSubtree(n->rChild);
  if (n->lChild != NULL)
    deleteSubtree(n->lChild);

  free(n);
}

static Node* create2Node(Keyval* kv, bool color) {
  Node* newInstancePtr = (Node*)malloc(sizeof(Node));
  if (newInstancePtr == NULL)
    return NULL;

  newInstancePtr->lChild = NULL;
  newInstancePtr->rChild = NULL;
  newInstancePtr->kv = kv;
  newInstancePtr->color = color;

  return newInstancePtr;
}

/* Si pas trouv, renvoit le noeud dans lequel la cl apparaitrait */
static Node* getNodeInSubtree(Node* n, int k) {
  if (k < n->kv->key) {
    if (n->lChild != NULL) return getNodeInSubtree(n->lChild, k);
    else                   return NULL;
  }
  else if (k > n->kv->key) {
    if (n->rChild != NULL) return getNodeInSubtree(n->rChild, k);
    else                   return NULL;
  }

  return n;
}

static Node* getNodeByKey(Tree* t, int k) {
  if (isEmpty(t))
    return NULL;

  return getNodeInSubtree(t->root, k);
}

static int cmpKeys(int a, int b) {
  return b - a;
}

static Node* putNode(Tree* t, Node* n, Keyval* kv) {
  /* -- First: insert the node at the right place --
     If the subtree where we put the node is empty, creating its root. */
  if (n == NULL) return create2Node(kv, RED);
  
  // Else record comparison result and increment comparisons counter
  int cmp = cmpKeys(kv->key, n->kv->key);
  t->nbCmp++;  
  
  /* Then if key in node sup. to the key we want to add, put the node on
     its left child */
  if      (cmp > 0) n->lChild = putNode(t, n->lChild, kv);
  
  /* Else if key in node inf. to the key we want to add, put the node on
     its right child */
  else if (cmp < 0) n->rChild = putNode(t, n->rChild, kv);
  
  // Else key in node equals key we want to add so replace existing value
  else strncpy(n->kv->value, kv->value, 20);

  
  /* -- Next: hold invariant true --
     #1: If red color not left-leaning for this node, rotate left */
  if (isRed(n->rChild) && !isRed(n->lChild)) n = rotateLeft(n);
  
  // #2: If red link succeeded by another one directly, rotate right
  if (isRed(n->lChild) && isRed(n->lChild->lChild)) n = rotateRight(n);
  
  // (case of two red links on the right never occures thx to #1)
  
  // #3: If both children are linked by red color to node, flip colors
  if (isRed(n->lChild) && isRed(n->rChild)) n = flipColors(n);

  // If a rotation occured, return the new root of the subtree
  return n;
}

static void inOrder(Node* n, Node* p) {
  int parKey = -1;
  if (p != NULL) parKey = p->kv->key;
  if(n != NULL) {
    inOrder(n->lChild, n);
    char* color = "BLACK\0";
    if (n->color == RED) color = "RED\0"; 
    if      (p == NULL)      printf("Root -- Key: %d, value='%s'\n", n->kv->key, n->kv->value);
    else if (n == p->lChild) printf("Left  child of '%d' -- Key: %d, value='%s' (incoming branch: %s)\n", parKey, n->kv->key, n->kv->value, color);
    else if (n == p->rChild) printf("Right child of '%d' -- Key: %d, value='%s' (incoming branch: %s)\n", parKey, n->kv->key, n->kv->value, color);
    inOrder(n->rChild, n);
  }
}


/******************************************************************************
*----------------------------- Accessible functions --------------------------*
******************************************************************************/
Tree* lcreateTree() {
  Tree* newInstancePtr = (Tree*)malloc(sizeof(Tree));
  if (newInstancePtr == NULL)
    return NULL;

  newInstancePtr->root = NULL;

  return newInstancePtr;
}

void ldeleteTree(Tree* t) {
  if (t->root != NULL)
    deleteSubtree(t->root);
  free(t);
}

/*int linsertNode(Tree* t, int key, char* value) {
  Keyval* kv = (Keyval*)malloc(sizeof(Keyval));
  kv->key = key;
  strncpy(kv->value, value, 20);

  t->nbCmp = 0;
  t->root = putNode(t, t->root, kv);

  return t->nbCmp;
}*/

int linsertNode(Tree* t, int key, char* value) {
  // -- First: creating key-value structure --
  Keyval* kv = (Keyval*)malloc(sizeof(Keyval));
  kv->key = key;
  strncpy(kv->value, value, 20);

  t->nbCmp = 0; // Reset comparisons counter   
  
  // -- Next: calling insertion on the entire tree --
  t->root = putNode(t, t->root, kv);
  return t->nbCmp;
}

void ldisplayNodeContent(Tree* t, int k) {
  Node* n = getNodeByKey(t, k);
  if (n == NULL)
    printf("No node within the tree recorded with this key.\n");
  else
    printf("Node found! Its value is '%s'.\n", n->kv->value);
}

void ldisplayTree(Tree* t) {
  inOrder(t->root, NULL);
}